package pi;

import java.io.*;
import java.math.*;

import mandala.util.*;

/**
   <p>Instance of this objet provides a mechanism to compute the greek number
   PI to an arbitrary precision.</p>
   
   <p>This class uses an object <code>interval</code>, an instance of the
   {@link IntInterval} to compute : <br>
   
   <blockquote>
   Sum(k = interval.start, k = interval.end,
   [1/(16^k)]*[4/(8k+1)-2/(8k+4)-1/(8k+5)-1/(8k+6)]
   </blockquote>
   
   <p>When <code>interval.start = 0</code>, the value <code>interval.end</code>
   is the precision of the desired value of PI. This formula is the BBP
   (Bailey-Borwein-Plouffe) formula (see <a
   href="http://www.cecm.sfu.ca/pi/pi.html">The Pi Pages</a> for more
   information.)</p>
   
   @author <a href="mailto:eipiequalsnoone@users.sf.net">eipi</a>
   @version $Revision: 1.4 $
   
   @see java.io.Serializable
   
*/
public class PiComputer implements Serializable {
    /**
       <p>Code for the {@link #rawCompute(IntInterval)} strategy.</p>
    */
    public static final String RAW = "raw".intern();

    /**
       <p>Code for the {@link #safeCompute(IntInterval)} strategy.</p>
    */
    public static final String SAFE = "safe".intern();
    /**
       <p>Code for the {@link #safeComputePerf(IntInterval)} strategy.</p>
    */
    public static final String SAFE_PERF = "safePerf".intern();
        
    public static Syslog syslog = new Syslog();
    static{
        syslog.DEBUG = false;
    }
        
    private static final int ROUND_MODE = BigDecimal.ROUND_HALF_EVEN;
                
    private BigDecimal ZERO = new BigDecimal("0");
    private BigDecimal ONE = new BigDecimal("1");
    private BigDecimal OPPOSITE_ONE  = new BigDecimal("-1");
    private BigDecimal TWO = new BigDecimal("2");
    private BigDecimal OPPOSITE_TWO = new BigDecimal("-2");
    //private static BigDecimal THREE = new BigDecimal("3"); // Unused
    private BigDecimal FOUR = new BigDecimal("4");
    private BigDecimal FIVE = new BigDecimal("5");
    private BigDecimal SIX = new BigDecimal("6");
    private BigDecimal HEIGHT = new BigDecimal("8");
    private BigDecimal OPPOSITE_HEIGHT = new BigDecimal("-8");
        
    private BigInteger SIXTEEN = new BigInteger("16");

    //    private int scale;

    public static void distribute(IntInterval[] intervals, int scale) {
        int nbPerIntInterval = scale / intervals.length;
        
        for (int i = 0; i < intervals.length; i++) {
            int start = i * nbPerIntInterval; 
            int end = ((i == intervals.length - 1) ? 
                       scale : 
                       start + nbPerIntInterval - 1);
            
            intervals[i] = new IntInterval(start, end);
        }
    }

                
    public PiComputer(int scale)
    {
        //      this.scale = scale;
                
        ZERO = ZERO.setScale(scale);
        ONE = ONE.setScale(scale);
        OPPOSITE_ONE = OPPOSITE_ONE.setScale(scale);
        TWO = TWO.setScale(scale);
        OPPOSITE_TWO = OPPOSITE_TWO.setScale(scale);
        //THREE = THREE.setScale(scale);
        FOUR = FOUR.setScale(scale);
        FIVE = FIVE.setScale(scale);
        SIX = SIX.setScale(scale);
        HEIGHT = HEIGHT.setScale(scale);
        OPPOSITE_HEIGHT = OPPOSITE_HEIGHT.setScale(scale);
    }
        
    /**
       <p>This is the method to call for the computation.</p>
           
       @param interval a <code>IntInterval</code> value
       @param strategy the strategy code to use
         
       @return a <code>Result</code> value
    */
    public Result compute(IntInterval interval, String strategy) {
        String strategyIntern = strategy.intern();
        if (strategyIntern == RAW) {
            return rawCompute(interval);
        }
        if (strategyIntern == SAFE) {
            return safeCompute(interval);
        }
        if (strategyIntern == SAFE_PERF) {
            return safeComputePerf(interval);
        }
                
        throw new IllegalArgumentException("Unknow strategy : " + strategy);
    }

    /**
       <p>Basic strategy.</p>
           
       <p>This method implements the most natural strategy which is not
       optimized at all.</p>
           
       @param interval a <code>IntInterval</code> value
       @return a <code>Result</code> value
    */
    public Result rawCompute(IntInterval interval) {
        long startChrono = System.currentTimeMillis();
        syslog.log("raw : computing from " + 
                   interval.start + " to " + interval.end);
                
        BigDecimal bd = ZERO;

        for (int k = interval.start; k <= interval.end; k++) {
            bd = bd.add(f(k));
        }
                                
        return  new Result(bd, 
                           System.currentTimeMillis() - startChrono);
        
    }

        
    /**
       This is the parameter used in <code>safeCompute(IntInterval)</code> and
       <code>safeComputePerf(IntInterval)</code>
           
       @see #safeCompute(IntInterval)
       @see #safeComputePerf(IntInterval)
    */
    public static final int LOOP_PARAMETER = 10;
        
    /**
       <p><em>Safe</em> non optimized strategy.</p>
           
       <p>This method implements the same strategy as the
       <code>rawCompute(IntInterval)</code> method but in a safest, cleaner,
       more responsive way. In fact, the current thread its <em>interrupt
       flag</em> every <code>LOOP_PARAMETER</code>. If the flag is set, the
       method abort the calculation.</p>
           
       @param interval an <code>IntInterval</code> value
       @return a <code>Result</code> value
    */
    public Result safeCompute(IntInterval interval) {
        long startChrono = System.currentTimeMillis();
        syslog.log("safe : computing from " + 
                   interval.start + " to " + interval.end);
                
        BigDecimal bd = ZERO;
                
        for (int k = interval.start; k <= interval.end; k++) {
            if ((k - interval.start) % LOOP_PARAMETER == 0){
                if (Thread.interrupted()){
                    syslog.log("Computation from " + interval.start + 
                               " to " + interval.end  +  
                               " : interrupted after " + 
                               (System.currentTimeMillis() - startChrono) + 
                               " ms.");
                    return null;
                }
            }
                        
            bd = bd.add(f(k));
        }
                
        return new Result(bd, 
                          System.currentTimeMillis() - startChrono);
    }

    /**
       <p><em>Safe</em> optimized strategy.</p>
           
       <p>This method implements the <em>safe</em> strategy but in a more
       efficient way. It does less test than the
       <code>safeCompute(IntInterval)</code> method, but require more
       computation. But the <code>rawCompute(IntInterval)</code> method which
       do not test interrupt() at all is still more efficient !</p>
           
       @param interval an <code>IntInterval</code> value
       @return a <code>Result</code> value
    */  
    public Result safeComputePerf(IntInterval interval) {
        long startChrono = System.currentTimeMillis();
        syslog.log("safePerf : computing from " + 
                   interval.start + " to " + interval.end);
                
        BigDecimal bd = ZERO;
        int ints = interval.end - interval.start;
                
        // loopNb = n - 1 loop of LOOP_PARAMETER elements. The last loop
        // contains less than LOOP_PARAMETER elements.
                
        int loopNb = (ints < LOOP_PARAMETER) ? 0 : ints / LOOP_PARAMETER;  
        int i, j;
                
        // Computing the first n - 1 loop
        for (i = 0; i < loopNb; i++) {
            if (Thread.interrupted()){
                syslog.log("Computation from " + interval.start + 
                           " to " + interval.end +  " : interrupted after " + 
                           (System.currentTimeMillis() - startChrono) + 
                           " ms.");
                return null;
            }
                        
                        
            for (j = 0; j < LOOP_PARAMETER; j++) {
                int k = interval.start + i * LOOP_PARAMETER + j;
                bd = bd.add(f(k));
            }
        }
                
        if (Thread.interrupted()){
            syslog.log("Computation from " + interval.start + " to " + 
                       interval.end +  " : interrupted after " + 
                       (System.currentTimeMillis() - startChrono) + " ms.");
            return null;
        }
                
        // Computing the last loop
        for (int k = interval.start + loopNb * LOOP_PARAMETER;
             k <= interval.end;
             k++){
            bd = bd.add(f(k));
        }
        
        return new Result(bd, 
                          System.currentTimeMillis() - startChrono);
    }
        
    private BigDecimal f(int k){
        BigDecimal K = new BigDecimal(k);
        BigDecimal HEIGHT_K = HEIGHT.multiply(K);
        BigDecimal FIRST = 
            ONE.divide(new BigDecimal(SIXTEEN.pow(k)), ROUND_MODE);
        BigDecimal SECOND = 
            FOUR.divide(HEIGHT_K.add(ONE), ROUND_MODE);
        BigDecimal THIRD = 
            OPPOSITE_TWO.divide(HEIGHT_K.add(FOUR), ROUND_MODE);
        BigDecimal FOURTH = 
            OPPOSITE_ONE.divide(HEIGHT_K.add(FIVE), ROUND_MODE);
        BigDecimal FIFTH = 
            OPPOSITE_ONE.divide(HEIGHT_K.add(SIX), ROUND_MODE);
                
        return FIRST.multiply(SECOND.add(THIRD.add(FOURTH.add(FIFTH))));
    }

    public static final String usage(Class usageClass) {
        return usage(usageClass, null);
    }

    public static final String usage(Class usageClass, String optionnal) {
        return "Usage : java " + 
            usageClass.getName() + 
            " Scale StrategyCode" + 
            ((optionnal != null) ? " " + optionnal : "") +
            "\nWhere StrategyCode can be : \n\"" + RAW + 
            "\", \"" + SAFE + "\", or \"" + SAFE_PERF + "\"";
    }
        
    public static void main(String[] args) {
        if (args.length != 2) {
            syslog.error(usage(PiComputer.class));
            System.exit(1);
        }

        long startChrono = System.currentTimeMillis();
                
        // Summin from 0 to interval.end to have the Pi value with a precision
        // of interval.end
        IntInterval interval = new IntInterval(0,
                                               Integer.parseInt(args[0]));

        // Multiplication of BigDecimal requires double scale !
        //              PiComputer computer = new PiComputer(interval.end * 2);
        PiComputer computer = new PiComputer(interval.end);

        Result result = computer.compute(interval, args[1]);
                
        logResult(interval,
                  1,
                  result,
                  System.currentTimeMillis() - startChrono);
                
    }

    public static void logResult(IntInterval interval,
                                 int nb,
                                 Result result,
                                 long totalTime) {
        // Truncate to interval.end + 2 - 1 decimals to avoid the beginning
        // '3.'  and scale start to 0 !
        String pi = result.bd.toString().substring(0, interval.end + 2); 
        long overheadTime = totalTime - result.computedTime;
        float overheadPercentage = 
            100 * (((float) totalTime / (float) result.computedTime) - 1);
        syslog.log("Sum(k = " + interval.start + ", k = " + interval.end + 
                   ",  [1/(16^k)]*[4/(8k+1)-2/(8k+4)-1/(8k+5)-1/(8k+6)] =\n" + 
                   pi + "\nSum CPU computed time : " + result.computedTime + " ms" +
                   "\nAverage CPU computed time :" + 
                   ((float) result.computedTime / (float) nb) + 
                   " ms" +
                   "\nReal time : " + totalTime + " ms" + 
                   "\nOverhead time : " + overheadTime + " ms (" + 
                   overheadPercentage + " %)");
    }

    public String toString(){
        return "PiComputer";
    }
        
    /**
       <p>This class wrap both the result of a computation and the time spent
       to compute the result.</p>
           
       @author <a href="mailto:eipiequalsnoone@users.sf.net">eipi</a>
       @version $Revision: 1.4 $
           
       @see Serializable
           
    */
    public static class Result implements Serializable {
        BigDecimal bd;
        long computedTime;

        public Result(BigDecimal bd, long computedTime) {
            this.bd = bd;
            this.computedTime = computedTime;
        }
    }
}               


